#include <px4_log.h>
#include <string.h>
#include "serial_port.h"
#include "slave_controller.h"
#include "master_controller.h"
#include "test_data_sender.h"


extern "C" __EXPORT int codisava_main(int argc, char *argv[]);

enum ExecutionMode {Slave, Master, SendData, None};
//Slave: receive messages from input serial port and dispatch them to either the motor controller or the next slave
//SendData: send test data through TELEM2 (used as output)

ExecutionMode ParseExecutionMode(int argc, char* argv[])
{
    for (int i= 0; i<argc; i++)
    {
        if (!strcmp(argv[i],"slave"))
            return ExecutionMode::Slave;
        if (!strcmp(argv[i],"send-data"))
            return ExecutionMode::SendData;
    }
    return ExecutionMode::None; //execution mode undefined
}


int codisava_main(int argc, char *argv[])
{
    SlaveController slaveController;
    TestDataSender testDataSender;

    ExecutionMode mode= ParseExecutionMode (argc, argv);

    if (mode==ExecutionMode::Slave)
        slaveController.Run(argc, argv);
    else if (mode == ExecutionMode::SendData)
        testDataSender.Run(argc, argv);

    fprintf(stderr, "Error: Undefined execution mode. Usage:\n\n");
    fprintf(stderr, "    -codisava slave id=[1,4] <no-feedback> <debug> <non-blocking>\n");
    fprintf(stderr, "    -codisava send-data <no-feedback> <debug> <non-blocking>\n");

    return OK;
}
