#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdio.h>
#include <termios.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <memory.h>


//ETX and STX of the frames
#define STX 2
#define ETX 3

#define MASTER_FRAME_SIZE 52
#define SLAVE_FRAME_SIZE 16
#define SLAVE_TIMEOUT_COUNT 10 //Will try 1 read per milisecond
#define SLAVE_TIMEOUT_TIME 100 //	Timeout waiting time in microseconds
int open_port(char *port)
{
	int fd;
	// Open serial port
	fd = open(port , O_RDWR | O_NOCTTY | O_NONBLOCK | O_NDELAY);
	// Check for Errors
	if (fd == -1)

	{
		printf("\nERROR: could not open '%s' serial port\n", port);
		return(-1);

	}else
		printf("\n'%s' port correctly opened\n", port);

	return fd;

}

int open_port_master(char *port)
{
	int fd;
	// Open serial port
	fd = open(port , O_RDWR);
	// Check for Errors
	if (fd == -1)

	{
		printf("\nERROR: could not open '%s' serial port\n", port);
		return(-1);

	}else
		printf("\n'%s' port correctly opened\n", port);

	return fd;

}

void config_port(int uart0_filestream)
{
	struct termios options;
		tcgetattr(uart0_filestream, &options);
		options.c_cflag = B115200 | CS8 | CLOCAL | CREAD;		//<Set baud rate
		options.c_iflag = IGNPAR;
		options.c_oflag = 0;
		options.c_lflag = 0;
		tcflush(uart0_filestream, TCIOFLUSH);
		tcsetattr(uart0_filestream, TCSANOW, &options);
}


void config_port_master(int uart0_filestream)
{
	struct termios options;
		tcgetattr(uart0_filestream, &options);
		options.c_cflag = B115200 | CS8 | CLOCAL | CREAD;		//<Set baud rate
		options.c_iflag = IGNPAR;
		options.c_oflag = 0;
		options.c_lflag = 0;
		sleep(1);
		tcflush(uart0_filestream, TCIOFLUSH);
		tcsetattr(uart0_filestream, TCSANOW, &options);
}


int read_port(int fd, char *buf, int len)
{

	int bytesRead = read(fd, buf, (unsigned) len);
	int num =0;

	if (bytesRead < 0)

	{
		//printf(stderr,"\nERROR: could not read port\n");
		return(-1);

	}else{

		//printf("\nMessage read correctly, bytes read: %d\n", bytesRead);

		for(int i=0; i<len; i++){
			if((int8_t)buf[i] < 0)
				num = (int8_t)buf[i] + 256;
			else
				num = (int8_t)buf[i];

       			printf("\nI(%d): %d\n",i,num);
			}
		}
	return bytesRead;


}


int read_port_frame(int fd, char *buf, int len, bool showPrints)
{
	int bytesReadTotal=0;
	int bytesRead=0;

	int num =0;

	bytesRead = read(fd, &buf[bytesReadTotal], (unsigned)len);
	if (bytesRead < 0)

	{
		//printf(stderr,"\nERROR: could not read port\n");
		return (-1);

	}else{

		bytesReadTotal += bytesRead;
		while(bytesReadTotal != len){
			bytesRead = read(fd, &buf[bytesReadTotal], (unsigned)len);
			bytesReadTotal += bytesRead;

		}

		if(showPrints){
		printf("\nMessage read correctly, bytes read in total: %d\n", bytesReadTotal);
		printf("\n");
		for(int i=0; i<bytesReadTotal; i++){
			if((int8_t)buf[i] < 0)
				num = (int8_t)buf[i] + 256;
			else
				num = (int8_t)buf[i];

       			printf("I(%d): %d\n",i,num);
			 }
		   }
			return bytesReadTotal;
		}

	return bytesReadTotal;
}


int write_port(int fd, char *buf, unsigned len)
{

	int bytesWritten = write(fd, buf, len);

	if (bytesWritten == -1)

	{
		//printf(stderr,"\nERROR: could not write in port\n");
		//close(fd);
		return(-1);

	}else
		//printf("\nMessage written correctly, bytes written: %d\n", bytesWritten);

	return bytesWritten;


}



unsigned char CalculateChecksumB(char* buf, int len)//len is 13 for slave's frames and 49 for master's frames
{
    unsigned char checksum = 0;
    //int numBytes = 13; //number of bytes used to calculate the checksum: 1(id) + 3 4-byte floats per slave
    //unsigned char* pByte= (unsigned char*) & msg.payload[0];

    for (int byte=1; byte<=len; byte++)
    {
        checksum = checksum xor buf[byte];
    }

    return checksum;
}


void setFramesToSlaves(char* bufMaster, char* bufSlave1, char* bufSlave2, char* bufSlave3, char* bufSlave4)
{
	////Slave 1////
	bufSlave1[0] = 2;
	bufSlave1[1] = bufMaster[1];
	bufSlave1[2] = bufMaster[2];
	bufSlave1[3] = bufMaster[3];
	bufSlave1[4] = bufMaster[4];
	bufSlave1[5] = bufMaster[5];
	bufSlave1[6] = bufMaster[6];
	bufSlave1[7] = bufMaster[7];
	bufSlave1[8] = bufMaster[8];
	bufSlave1[9] = bufMaster[9];
	bufSlave1[10] = bufMaster[10];
	bufSlave1[11] = bufMaster[11];
	bufSlave1[12] = bufMaster[12];
	bufSlave1[13] = bufMaster[13];
	bufSlave1[14] = CalculateChecksumB(bufSlave1, SLAVE_FRAME_SIZE-3);
	bufSlave1[15] = 3;

	////Slave 2////
	bufSlave2[0] = 2;
	bufSlave2[1] = bufMaster[1];
	bufSlave2[2] = bufMaster[14];
	bufSlave2[3] = bufMaster[15];
	bufSlave2[4] = bufMaster[16];
	bufSlave2[5] = bufMaster[17];
	bufSlave2[6] = bufMaster[18];
	bufSlave2[7] = bufMaster[19];
	bufSlave2[8] = bufMaster[20];
	bufSlave2[9] = bufMaster[21];
	bufSlave2[10] = bufMaster[22];
	bufSlave2[11] = bufMaster[23];
	bufSlave2[12] = bufMaster[24];
	bufSlave2[13] = bufMaster[25];
	bufSlave2[14] = CalculateChecksumB(bufSlave2, SLAVE_FRAME_SIZE-3);
	bufSlave2[15] = 3;

	////Slave 3////
	bufSlave3[0] = 2;
	bufSlave3[1] = bufMaster[1];
	bufSlave3[2] = bufMaster[26];
	bufSlave3[3] = bufMaster[27];
  bufSlave3[4] = bufMaster[28];
	bufSlave3[5] = bufMaster[29];
	bufSlave3[6] = bufMaster[30];
	bufSlave3[7] = bufMaster[31];
	bufSlave3[8] = bufMaster[32];
	bufSlave3[9] = bufMaster[33];
	bufSlave3[10] = bufMaster[34];
	bufSlave3[11] = bufMaster[35];
	bufSlave3[12] = bufMaster[36];
	bufSlave3[13] = bufMaster[37];
	bufSlave3[14] = CalculateChecksumB(bufSlave3, SLAVE_FRAME_SIZE-3);
	bufSlave3[15] = 3;

	////Slave4////
	bufSlave4[0] = 2;
	bufSlave4[1] = bufMaster[1];
	bufSlave4[2] = bufMaster[38];
	bufSlave4[3] = bufMaster[39];
	bufSlave4[4] = bufMaster[40];
	bufSlave4[5] = bufMaster[41];
	bufSlave4[6] = bufMaster[42];
	bufSlave4[7] = bufMaster[43];
	bufSlave4[8] = bufMaster[44];
	bufSlave4[9] = bufMaster[45];
	bufSlave4[10] = bufMaster[46];
	bufSlave4[11] = bufMaster[47];
	bufSlave4[12] = bufMaster[48];
	bufSlave4[13] = bufMaster[49];
	bufSlave4[14] = CalculateChecksumB(bufSlave4, SLAVE_FRAME_SIZE-3);
	bufSlave4[15] = 3;
}


void setFrameToMaster(char* bufMaster, char* bufSlave1, char* bufSlave2, char* bufSlave3, char* bufSlave4, int timeoutID)
{
	//STX and ID
	bufMaster[0] = 2;
	bufMaster[1] = timeoutID;

	/////////////////////////////PAYLOAD////////////////////////////////
	/////Slave 1 payload/////
	//pitch
	bufMaster[2] = bufSlave1[2] ;
	bufMaster[3] = bufSlave1[3] ;
	bufMaster[4] = bufSlave1[4] ;
	bufMaster[5] = bufSlave1[5] ;
	//roll
	bufMaster[6] = bufSlave1[6] ;
	bufMaster[7] = bufSlave1[7] ;
	bufMaster[8] = bufSlave1[8] ;
	bufMaster[9] = bufSlave1[9] ;
	//thrust
  bufMaster[10] = bufSlave1[10] ;
	bufMaster[11] = bufSlave1[11] ;
	bufMaster[12] = bufSlave1[12] ;
	bufMaster[13] = bufSlave1[13] ;
	/////////////////////////
	/////////////////////////
	////Slave 2 payload//////
	//pitch
	bufMaster[14] = bufSlave2[2] ;
	bufMaster[15] = bufSlave2[3] ;
	bufMaster[16] = bufSlave2[4] ;
	bufMaster[17] = bufSlave2[5] ;
	//roll
	bufMaster[18] = bufSlave2[6] ;
	bufMaster[19] = bufSlave2[7] ;
	bufMaster[20] = bufSlave2[8] ;
	bufMaster[21] = bufSlave2[9] ;
	//thrust
  bufMaster[22] = bufSlave2[10] ;
	bufMaster[23] = bufSlave2[11] ;
	bufMaster[24] = bufSlave2[12] ;
	bufMaster[25] = bufSlave2[13] ;
	/////////////////////////
	/////////////////////////
	////Slave 3 payload//////
	//pitch
	bufMaster[26] = bufSlave3[2];
	bufMaster[27] = bufSlave3[3];
	bufMaster[28] = bufSlave3[4];
	bufMaster[29] = bufSlave3[5];
	//roll
	bufMaster[30] = bufSlave3[6];
	bufMaster[31] = bufSlave3[7];
	bufMaster[32] = bufSlave3[8];
	bufMaster[33] = bufSlave3[9];
	//thrust
	bufMaster[34] = bufSlave3[10];
	bufMaster[35] = bufSlave3[11];
	bufMaster[36] = bufSlave3[12];
	bufMaster[37] = bufSlave3[13];
	////////////////////////
	////////////////////////
	////Slave 4 payload/////
	//pitch
	bufMaster[38] = bufSlave4[2];
	bufMaster[39] = bufSlave4[3];
	bufMaster[40] = bufSlave4[4];
	bufMaster[41] = bufSlave4[5];
	//roll
	bufMaster[42] = bufSlave4[6];
	bufMaster[43] = bufSlave4[7];
	bufMaster[44] = bufSlave4[8];
	bufMaster[45] = bufSlave4[9];
	//thrust
	bufMaster[46] = bufSlave4[10];
	bufMaster[47] = bufSlave4[11];
	bufMaster[48] = bufSlave4[12];
	bufMaster[49] = bufSlave4[13];
	/////////////////////////
	////////////////////////////////////////////////////////////////////

	//ETX and Checksum
	bufMaster[50] = CalculateChecksumB(bufMaster, MASTER_FRAME_SIZE-3);
	bufMaster[51] = 3;
}


bool isValid(char* buf, int len)
{
    //first, the easy part: check the beginning and the end of the message
    if (buf[0] != STX || buf[len-1] != ETX)
		{
		//	PX4_INFO("STX and ETX are incorrect");
			return false;
		}
    //now, the checksum
    if (CalculateChecksumB(buf, len-3) == buf[len-2])
		{
		//PX4_INFO("CORRECT MESSAGE");
    	return true;
		}
		//PX4_INFO("INVALID CHECKSUM");
    return false;
}


int resynchronize(char *buf, int len)
{

		int bytesCorrect=0, i=0;
		bool stxFound=false;
		//Search STX in the incorrect buffer
		while(!stxFound){
				if (buf[i] == STX){
					  buf[0]=buf[i];
						bytesCorrect++;
						stxFound = true;
				}
	    	i++;
		}

		//Save the correct part of the buffer
		while(i<len){
				buf[bytesCorrect] = buf[i];
				bytesCorrect++;
				i++;
		}


 	return bytesCorrect;

}


int main(int argc, char *argv[])
{

	//sleep(3);

	//Input-output counter variables
	int bytesWritten, bytesRead;
	int framesSent1 =0, framesSent2 =0, framesReceived1 =0, framesReceived2 =0;
	int framesSent3 =0, framesSent4 =0, framesReceived3 =0, framesReceived4 =0;

	//variables for resynchro and frame validation
  syncBytes = 0;


	//Open the ports
	int fdSlave1 = open_port("/dev/ttyUSB0");
	int fdSlave2 = open_port("/dev/ttyUSB1");
	int fdSlave3 = open_port("/dev/ttyUSB2");
	int fdSlave4 = open_port("/dev/ttyUSB3");
	int fdMaster = open_port_master("/dev/ttyS0");

	//Port configuration
	config_port(fdSlave1);
	config_port(fdSlave2);
	config_port(fdSlave3);
	config_port(fdSlave4);
	config_port_master(fdMaster);

	//Create the buffers
	char bufMaster[MASTER_FRAME_SIZE];
	char bufSlave1[SLAVE_FRAME_SIZE];
	char bufSlave2[SLAVE_FRAME_SIZE];
	char bufSlave3[SLAVE_FRAME_SIZE];
	char bufSlave4[SLAVE_FRAME_SIZE];

	//Init Buffers
	for(int i=0; i< MASTER_FRAME_SIZE; i++)  bufMaster[i] = 0;
	for(int i=0; i< SLAVE_FRAME_SIZE; i++){
			bufSlave1[i] = 0;
			bufSlave2[i] = 0;
			bufSlave3[i] = 0;
			bufSlave4[i] = 0;
	}

	//Read counter for timeout and ID to warn the master
	int timeoutCount;
	int timeoutID;

	//Variables to control monitorizing freq
	int showPrintsCount;
	int showPrints;

	if (fdSlave1 == -1 || fdSlave2 == -1 || fdSlave3 == -1 || fdSlave4 == -1 || fdMaster == -1)

	{
		close(fdSlave1);
		close(fdSlave2);
		close(fdSlave3);
		close(fdSlave4);
		close(fdMaster);
		return(-1);

	}else{

		//bool desync = false;
		while(1)
		{

			showPrintsCount++;

			if(showPrintsCount > 10000){
				showPrints = true;
				showPrintsCount = 0;
			}else if(showPrints) showPrints=false;

			//Set ID to 0 at first
			timeoutID = 0;
			//if (desync) printf("Desynchronization occurred\n");
			//Receive the frame from the master and prepare it for the slaves
			read_port_frame(fdMaster, &bufMaster[syncBytes], MASTER_FRAME_SIZE-syncBytes, showPrints); //master
			timeoutID = bufMaster[1];
			//Check if the message is valid
			if(isValid(bufMaster, MASTER_FRAME_SIZE)){
				syncBytes=0;
				setFramesToSlaves(bufMaster, bufSlave1, bufSlave2, bufSlave3, bufSlave4);
			}else{
				incorrectCount++;
				if(incorrectCount >= 1){
						printf("Desynchronization occurred\n");
						syncBytes = resynchronize(bufMaster, MASTER_FRAME_SIZE);
						incorrectCount=0;
				}

			}

				//Send the frame to the slaves
				bytesWritten = write_port(fdSlave1, bufSlave1, SLAVE_FRAME_SIZE);//slave1
				framesSent1++;
				bytesWritten = write_port(fdSlave2, bufSlave2, SLAVE_FRAME_SIZE);//slave2
				framesSent2++;
				bytesWritten = write_port(fdSlave3, bufSlave3, SLAVE_FRAME_SIZE);//slave3
				framesSent3++;
				bytesWritten = write_port(fdSlave4, bufSlave4, SLAVE_FRAME_SIZE);//slave4
				framesSent4++;
				//Receive the feedback from the slaves

										////Slave1////
				timeoutCount = 0;
				bytesRead = read_port_frame(fdSlave1, bufSlave1, SLAVE_FRAME_SIZE, showPrints); //Try to read
				if(bytesRead < 0){
					printf("Could not read port, triying to read...\n");
					while(bytesRead < 0 && timeoutCount < SLAVE_TIMEOUT_COUNT){ //Try to read X times
						bytesRead = read_port_frame(fdSlave1, bufSlave1, SLAVE_FRAME_SIZE, showPrints);
						timeoutCount++;
						usleep(SLAVE_TIMEOUT_TIME);
						}
						if(bytesRead < 0){
							printf("Timeout occurred in Slave 1\n");
							timeoutID |= (1u << 4); //Change ID byte to inform the master
						}
						else
							framesReceived1++;
					}else
						framesReceived1++;

										////Slave2////
				timeoutCount = 0;
				bytesRead = read_port_frame(fdSlave2, bufSlave2, SLAVE_FRAME_SIZE, showPrints); //Try to read
				if(bytesRead < 0){
					printf("Could not read port, triying to read...\n");
					while(bytesRead < 0 && timeoutCount < SLAVE_TIMEOUT_COUNT){ //Try to read X times
						bytesRead = read_port_frame(fdSlave2, bufSlave2, SLAVE_FRAME_SIZE, showPrints);
						timeoutCount++;
						usleep(SLAVE_TIMEOUT_TIME);
						}
						if(bytesRead < 0){
						  printf("Timeout occurred in Slave 2\n");
							timeoutID |= (1u << 5); //Change ID byte to inform the master
						}
						else
							framesReceived2++;
					}else
						framesReceived2++;


									   ////Slave3////
				timeoutCount = 0;
				bytesRead = read_port_frame(fdSlave3, bufSlave3, SLAVE_FRAME_SIZE, showPrints); //Try to read
				if(bytesRead < 0){
			    printf("Could not read port, triying to read...\n");
					while(bytesRead < 0 && timeoutCount < SLAVE_TIMEOUT_COUNT){ //Try to read X times
						bytesRead = read_port_frame(fdSlave3, bufSlave3, SLAVE_FRAME_SIZE, showPrints);
						timeoutCount++;
						usleep(SLAVE_TIMEOUT_TIME);
						}
						if(bytesRead < 0){
						  printf("Timeout occurred in Slave 3\n");
							timeoutID |= (1u << 6); //Change ID byte to inform the master
						}
						else
							framesReceived3++;
					}else
						framesReceived3++;


									 ////Slave4////
				timeoutCount = 0;
				bytesRead = read_port_frame(fdSlave4, bufSlave4, SLAVE_FRAME_SIZE, showPrints); //Try to read
				if(bytesRead < 0){
					printf("Could not read port, triying to read...\n");
					while(bytesRead < 0 && timeoutCount < SLAVE_TIMEOUT_COUNT){ //Try to read X times
						bytesRead = read_port_frame(fdSlave4, bufSlave4, SLAVE_FRAME_SIZE, showPrints);
						timeoutCount++;
						usleep(SLAVE_TIMEOUT_TIME);
						}
						if(bytesRead < 0){
						  printf("Timeout occurred in Slave 4\n");
							timeoutID |= (1u << 7); //Change ID byte to inform the master
						}
						else
							framesReceived4++;
					}else
						framesReceived4++;

				//Re-construct the frame to send the feedback received from the slaves to the master
				setFrameToMaster(bufMaster, bufSlave1, bufSlave2, bufSlave3, bufSlave4, timeoutID);
				bytesWritten = write_port(fdMaster, bufMaster, MASTER_FRAME_SIZE);

				//Print the results of the communication
				if(showPrints) printf("Frames SENT: slave1-> %d, slave2-> %d, slave3-> %d, slave4-> %d\n", framesSent1, framesSent2, framesSent3, framesSent4);
				if(showPrints) printf("Frames RECEIVED: slave1-> %d, slave2-> %d, slave3-> %d, slave4-> %d\n", framesReceived1, framesReceived2, framesReceived3, framesReceived4);

				//Sleep 20 miliseconds
				usleep(20000);


		}
	}


	close(fdSlave1);
	close(fdSlave2);
	close(fdSlave3);
	close(fdSlave4);
	close(fdMaster);
	return 0;
}
